#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Implementation of SubclassIdentifiable interface-like class."""

from collections.abc import Hashable
from typing import Any, Optional

from design_pytterns.errors import InvalidClassIdError, UnhashableClassIdError


class SubclassIdentifiable():
    """
    Inject ``CLASS_ID`` class attribute to all the subclasses of the base.

    The subclasses of the base that inherits from this class require the
    ``class_id`` keyword argument in the class definition with a `Hashable` not
    `None` value.

    .. versionchanged:: 0.4.0

        - raise `UnhashableClassIdError` instead of `TypeError`.
        - raise `InvalidClassIdError` instead of `ValueError`.

    Raises
    ------
    UnhashableClassIdError
        If `class_id` is not `Hashable`.
    InvalidClassIdError
        If `class_id` is None for any of the subclasses.

    Examples
    --------
    >>> from design_pytterns.interfaces import SubclassIdentifiable
    >>>
    >>> class MyBase(SubclassIdentifiable):
    ...    pass
    ...
    >>> class MyIdentifiableSubclass(MyBase, class_id=1):
    ...    pass
    ...
    >>> MyIdentifiableSubclass.CLASS_ID
    1

    """

    def __init_subclass__(
            cls, class_id: Optional[Hashable] = None, **kwargs: Any
    ) -> None:
        super().__init_subclass__(**kwargs)

        if not isinstance(class_id, Hashable):
            raise UnhashableClassIdError(class_id)

        if SubclassIdentifiable.mro()[0] not in cls.__bases__:
            if class_id is None:
                raise InvalidClassIdError()

        cls.CLASS_ID = class_id



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Package with implementations of Singleton-pattern."""

from .singleton import Singleton
from .singleton_meta import SingletonMeta

__all__ = ['Singleton', 'SingletonMeta']



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Unhashable class identifier error."""

from typing import Any

from .design_pyttern_error import DesignPytternError


class UnhashableClassIdError(DesignPytternError):
    """
    Raised when the ``CLASS_ID`` of a `SubclassIdentifiable` is not hashable.

    Parameters
    ----------
    class_id : Any
        Unhashable value that raised the exception.

    """

    def __init__(self, class_id: Any):
        super().__init__(
            message=f'unhashable type: {type(class_id).__name__}'
        )



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Generic Event Publisher-pattern."""

from typing import Any, Hashable, Sequence

from .publisher import Publisher
from .subscriber import Subscriber


class EventPublisher():
    """
    Event based Publisher-pattern.

    Parameters
    ----------
    events: Sequence
        The sequence that contains all the possible hashable events

    """

    def __init__(self, events: Sequence[Hashable]):
        self._publishers = {event: Publisher() for event in events}

    def subscribe(self, event: Hashable, subscriber: Subscriber):
        """
        Register a new subscriber for an event.

        If the subscriber instance is already registered for the event the
        subscription is discarded.

        Parameters
        ----------
        event: Hashable
            The event which the subscriber registers to.
        subscriber : Subscriber
            The subscriber instance to register.

        """
        self._publishers[event].subscribe(subscriber)

    def unsubscribe(self, event: Hashable, subscriber: Subscriber):
        """
        Deregister the subscriber from an event.

        If the subscriber instance is unknown for the given event, the
        unsubscription is discarded.

        Parameters
        ----------
        event: Hashable
            The event which the subscriber deregister from.
        subscriber : Subscriber
            The subscriber instance to deregister.

        """
        self._publishers[event].unsubscribe(subscriber)

    def notify(self, event: Hashable, *args: Any, **kwargs: Any):
        """
        Execute the ``update`` method of all the event subscribers.

        Parameters
        ----------
        event : Hashable
            The event which all its subscribers are notified.
        *args : Any
            Variable length contextual arguments.
        **kwargs : Any
            Contextual keyword arguments.

        """
        self._publishers[event].notify(*args, **kwargs)



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Invalid class identifier error."""

from .design_pyttern_error import DesignPytternError


class InvalidClassIdError(DesignPytternError):
    """Raised when the ``CLASS_ID`` of a `SubclassIdentifiable` is `None`."""

    def __init__(self):
        super().__init__(message="invalid subclass identifier: 'None'")



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Singleton metaclass."""

from typing import Any, Dict, Tuple


class SingletonMeta(type):
    """Metaclass implementation of Singleton-pattern."""

    def __init__(
            cls, name: str, bases: Tuple[type, ...], mmbs: Dict[str, Any]
    ) -> None:
        super().__init__(name, bases, mmbs)
        cls._instance = None

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        """Invoke ``__call__`` method only if global instance is `None`."""
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Generic Publisher-pattern."""

from typing import Any, List

from .subscriber import Subscriber


class Publisher():
    """Generic Publisher-pattern."""

    def __init__(self):
        self._subscribers: List[Subscriber] = []

    def subscribe(self, subscriber: Subscriber):
        """
        Register a new subscriber.

        If the subscriber instance is already registered the subscription is
        discarded.

        Parameters
        ----------
        subscriber : Subscriber
            The subscriber instance to register.

        """
        if subscriber not in self._subscribers:
            self._subscribers.append(subscriber)

    def unsubscribe(self, subscriber: Subscriber):
        """
        Deregister the subscriber.

        If the subscriber instance is unknown, the unsubscription is discarded.

        Parameters
        ----------
        subscriber : Subscriber
            The subscriber instance to deregister.

        """
        if subscriber in self._subscribers:
            self._subscribers.remove(subscriber)

    def notify(self, *args: Any, **kwargs: Any):
        """
        Execute the ``update`` method of all the registered subscribers.

        Parameters
        ----------
        *args : Any
            Variable length contextual arguments.
        **kwargs : Any
            Contextual keyword arguments.

        """
        for subscriber in self._subscribers:
            subscriber.update(*args, **kwargs)



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Generic Factory-pattern."""

import logging
from typing import Any, Hashable, MutableMapping, Optional

from design_pytterns.errors import UnregisteredClassIdError


class Factory():
    """
    Generic Factory-pattern based on hashable id class registration.

    Parameters
    ----------
    registered_classes: dict, optional
        A dictionary that maps the hashable identifiers with their class types.

    """

    __LOGGER = logging.getLogger(__name__)

    def __init__(
            self,
            registered_classes: Optional[MutableMapping[Hashable, type]] = None
    ) -> None:
        if registered_classes is None:
            self._registered_classes = {}
        else:
            self._registered_classes = registered_classes

    def register_class(self, class_id: Hashable, class_type: type) -> None:
        """
        Register a class by a hashable identifier.

        Parameters
        ----------
        class_id : Hashable
            The class identifer.
        class_type : type
            The class type.

        Warnings
        --------
        Overwrite any previously mapped type if its class id is already in use
        and log a warning about the replacement.

        """
        if class_id in self._registered_classes:
            self.__LOGGER.warning(
                'class id %s is used by %s, replacing it with %s',
                repr(class_id),
                repr(self._registered_classes[class_id]),
                repr(class_type)
            )

        self._registered_classes[class_id] = class_type

    def create(self, class_id: Hashable, *args: Any, **kwargs: Any) -> Any:
        """
        Create a class instance given its identifier and constructor arguments.

        .. versionchanged:: 0.2.0 raises `UnregisteredClassIdError` instead of
            `TypeError`.

        Parameters
        ----------
        class_id : Hashable
            The class identifer.
        *args : Any
            Variable length arguments of the class constructor associated to
            the value of `class_id`.
        **kwargs: Any
            Keyword arguments of class constructor associated to the value of
            `class_id`.

        Returns
        -------
        Any
            Initialized class instance associated to the value of `class_id`.

        Raises
        ------
        UnregisteredClassIdError
            If `class_id` is unknown.

        """
        if class_id in self._registered_classes:
            return self._registered_classes[class_id](*args, **kwargs)

        raise UnregisteredClassIdError(class_id)



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Unregistered class id error."""

from collections.abc import Hashable

from .design_pyttern_error import DesignPytternError


class UnregisteredClassIdError(DesignPytternError):
    """
    Raised when a factory creates an instance of an unknown class id.

    Parameters
    ----------
    class_id : Hashable
        Unknown hashable that raised the exception.

    """

    def __init__(self, class_id: Hashable) -> None:
        super().__init__(
            class_id, message=f'unregistered class id: {repr(class_id)}'
        )



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Generic package error."""

from typing import Any


class DesignPytternError(Exception):
    """
    Base class for all design_patterns exceptions.

    Parameters
    ----------
    *args : Any
        Variable length arguments.
    **kwargs : Any
        Arbitrary keyword arguments.

    Attributes
    ----------
    message : str
        A descriptive sentence explaining the error.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.message = kwargs.pop('message', '')

        super().__init__(*args, **kwargs)



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later


"""Python Design Patterns."""

from ._about import (
    __author__,
    __author_email__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__
)

__all__ = [
    '__author__',
    '__author_email__',
    '__description__',
    '__license__',
    '__title__',
    '__url__',
    '__version__'
]



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Subclass Factory."""

from typing import Type

from design_pytterns._helpers import ConcreteSubclassRegister
from design_pytterns.interfaces import SubclassIdentifiable

from .factory import Factory


class SubclassFactory(Factory):
    """
    Automatic concrete sublass registration factory.

    Factory that analyzes the subclass hierarchy of the given base class that
    inherits from `SubclassIdentifiable` and registers all of the concrete
    subclasses with their associated class id automatically.

    Parameters
    ----------
    base_class: SubclassIdentifiable
        The base class that servers as the starting point of the subclass
        hierarchy analysis.

    """

    def __init__(self, base_class: Type[SubclassIdentifiable]):
        super().__init__(
            ConcreteSubclassRegister(base_class).registered_classes
        )



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Unidentifiable subclass error."""

from .design_pyttern_error import DesignPytternError


class UnidentifiableSubclassError(DesignPytternError):
    """
    Raised when the base class of a subclass register is not identifiable.

    Parameters
    ----------
    base_class : type
        Unidentifiable type that raised the exception.

    """

    def __init__(self, base_class: type):
        super().__init__(
            message=f'unidentifiable base class: {repr(base_class.__name__)}'
        )



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Subpackage with the definition of the metadata of Design-Pyttern package."""

__author__ = 'Alchemy-Meister'
__author_email__ = 'meister.alchemy@gmail.com'
__description__ = 'Implementation of Basic Python Design Patters'
__license__ = 'GPL-3.0-or-later'
__title__ = 'design-pytterns'
__url__ = 'https://github.com/Alchemy-Meister/python-design-patterns'
__version__ = '0.4.0'



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Package with interface-like classes."""

from .subclass_identifiable import SubclassIdentifiable

__all__ = ['SubclassIdentifiable']



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Concrete Subclass Register."""

from collections.abc import Hashable
from inspect import isabstract
import logging
from typing import MutableMapping, Type

from design_pytterns.errors import UnidentifiableSubclassError
from design_pytterns.interfaces import SubclassIdentifiable


class ConcreteSubclassRegister():
    """
    Concrete subclass register.

    Analyze the subclass hierarchy of the given base class that
    inherits from `SubclassIdentifiable` and registers all of the concrete
    subclasses with their associated class id automatically.

    Parameters
    ----------
    base_class: Type[SubclassIdentifiable]
        The base class that servers as the starting point of the subclass
        hierarchy analysis.

    Raises
    ------
    UnidentifiableSubclassError
        If `base_class` does not inherit from SubclassIdentifiable.

    .. versionchanged:: 0.3.0 raises `UnidentifiableSubclassError` instead of
            `TypeError`.

    """

    __LOGGER = logging.getLogger(__name__)

    def __init__(self, base_class: Type[SubclassIdentifiable]) -> None:
        if not issubclass(base_class, SubclassIdentifiable):
            raise UnidentifiableSubclassError(base_class)

        self.registered_classes = self._register_subclasses(base_class)

    def _register_subclasses(
            self, base_class: Type[SubclassIdentifiable]
    ) -> MutableMapping[Hashable, Type[SubclassIdentifiable]]:
        registered_classes = {}

        current_class = base_class
        pending_classes = current_class.__subclasses__()

        while pending_classes:
            current_class = pending_classes.pop(0)
            if current_class not in registered_classes.values():
                if not isabstract(current_class):
                    if current_class.CLASS_ID in registered_classes.keys():
                        self.__LOGGER.warning(
                            'class id %s is used by %s, replacing it with %s',
                            repr(current_class.CLASS_ID),
                            repr(registered_classes[current_class.CLASS_ID]),
                            repr(current_class)
                        )

                    registered_classes.update(
                        {current_class.CLASS_ID: current_class}
                    )

                pending_classes = (
                    current_class.__subclasses__() + pending_classes
                )

        return registered_classes



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Generic Subscriber-pattern."""

from abc import ABC, abstractmethod
from typing import Any


class Subscriber(ABC):
    """Abstract subscriber. Define the generic ``update`` method."""

    @abstractmethod
    def update(self, *args: Any, **kwargs: Any):
        """
        Abstract update method.

        Parameters
        ----------
        *args : Any
            Variable length arguments.
        **kwargs : Any
            Arbitrary keyword arguments.

        """



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Package with implementations of Factory-patterns."""

from .factory import Factory
from .subclass_factory import SubclassFactory

__all__ = ['Factory', 'SubclassFactory']



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Package distribution based on setuptools."""

from os.path import abspath, dirname, join
from setuptools import setup, find_packages

about = {}

with open(
        join(dirname(abspath(__file__)), 'src/design_pytterns/_about.py'), 'r'
) as about_file:
    exec(about_file.read(), about)  # nosec

with open('README.md', 'r') as readme_file:
    readme = readme_file.read()

install_requires = []

tests_require = [
    'coverage',
    'pytest',
    'pytest-cov'
]

extras_require = {'tests': tests_require}

setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__url__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords="Design Patterns, Singleton, Factory, Observer",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: '
        'GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        "Operating System :: OS Independent"
    ],
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require
)



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Package with the implementations of Observer-pattern."""

from .event_publisher import EventPublisher
from .publisher import Publisher
from .subscriber import Subscriber

__all__ = ['EventPublisher', 'Publisher', 'Subscriber']



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Subpackage with internal helper classes."""

from .concrete_subclass_register import ConcreteSubclassRegister

__all__ = ['ConcreteSubclassRegister']



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Subpackage with the exception definitions."""

from .design_pyttern_error import DesignPytternError
from .invalid_class_id_error import InvalidClassIdError
from .unhashable_class_id_error import UnhashableClassIdError
from .unidentifiable_subclass_error import UnidentifiableSubclassError
from .unregistered_class_id_error import UnregisteredClassIdError

__all__ = [
    'DesignPytternError',
    'InvalidClassIdError',
    'UnhashableClassIdError',
    'UnidentifiableSubclassError',
    'UnregisteredClassIdError'
]



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2020 Alchemy-Meister
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Generic Singleton-pattern."""

from .singleton_meta import SingletonMeta


class Singleton(metaclass=SingletonMeta):
    """
    Class implementation of Singleton-pattern.

    This implementation of the Singleton pattern calls the ``__init__``
    method of the subclass once, for the initialization of the global
    instance. In the following calls to the constructor, the ``__init__``
    method is not executed, ignoring any argument passed to it and returning
    the very first created instance instead.

    Examples
    --------
    >>> from design_pytterns.singleton import Singleton
    >>>
    >>> class MyNumber(Singleton):
    ...     def __init__(self, number):
    ...         self.number = number
    ...     def __str__(self):
    ...         return self.number
    ...
    >>> test_1 = MyNumber(5)
    >>> test_2 = MyNumber()  # No TypeError is raised
    >>> test_3 = MyNumber(10)
    >>>
    >>> id(test_1), test_1
    (139703749013216, 5)
    >>> id(test_2), test_2
    (139703749013216, 5)
    >>> id(test_3), test_3
    (139703749013216, 5)

    """
